#ifndef OnX_GuiViewer_h
#define OnX_GuiViewer_h

#include <Lib/smanip.h>
#include <Lib/Arguments.h>

// Inventor :
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoSeparator.h>

// HEPVis :
#include <HEPVis/SbPainterPS.h>
#include <HEPVis/SbString.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoViewportRegion.h>
#include <HEPVis/nodes/SoImageWriter.h>
#include <HEPVis/nodes/SoImage.h>
#include <HEPVis/nodekits/SoPage.h>
#include <HEPVis/nodekits/SoRegion.h>
#include <HEPVis/nodekits/SoDisplayRegion.h>
#include <HEPVis/nodekits/SoImageRegion.h>
#include <HEPVis/actions/SoPainterAction.h>
#include <HEPVis/actions/SoGL2PSAction.h>

template <class SoGuiViewer> 
class GuiPageSetRegionCallback : public virtual SbSetRegionCallback {
public:
  GuiPageSetRegionCallback(SoGuiViewer& aViewer):fViewer(aViewer){}
  virtual void setCamera(SoCamera* aCamera) {
    if(aCamera==fViewer.getCamera()) return; //done.
    fViewer.setCamera(aCamera);
  }
private:
  SoGuiViewer& fViewer;
};

template <class SoGuiViewer> 
inline void GuiViewer_initialize(SoGuiViewer& aThis) {
  const SbViewportRegion& vpRegion = aThis.getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  aThis.setGLRenderAction(action);

  SoPage* soPage = new SoPage;
  soPage->ref();
  // SetRegionCallback object deleted by the SoPage.
  soPage->setSetRegionCallback
    (new GuiPageSetRegionCallback<SoGuiViewer>(aThis));
    
  aThis.setSceneGraph(soPage);
    
  soPage->highlightRegion(soPage->getRootRegion(0));

  // Disconnect headlight since all regions have their own light.
  // (Headlight will add a directinal light, then scene will
  // be more bright).
  // OSF1/cxx needs the "this->".
  aThis.setHeadlight(FALSE);
}

template <class SoGuiViewer> 
inline SoPage* GuiViewer_SoPage(SoGuiViewer& aThis) {
  SoNode* node = aThis.getSceneGraph();
  if(!node) return 0;        
  if(node->isOfType(SoPage::getClassTypeId())==FALSE) return 0;
  return (SoPage*)node;
}

template <class SoGuiViewer> 
inline bool GuiViewer_finalize(SoGuiViewer& aThis) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  soPage->setSetRegionCallback(0);
  soPage->unref();
  return true;
}

template <class SoGuiViewer> 
inline bool GuiViewer_writeGIF(SoGuiViewer& aThis,const std::string& aFile) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;

  SoImageWriter* imageWriter = soPage->getImageWriter();
  imageWriter->fileName.setValue(aFile.c_str());
  imageWriter->format.setValue(SoImageWriter::GIF);
  imageWriter->enable();

  SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
  viewportRegion->highlighted.setValue(FALSE);
  aThis.render();
  viewportRegion->highlighted.setValue(TRUE);

  imageWriter->disable();
  if(imageWriter->getStatus()) {
    SoDebugError::postInfo("","%s produced.",
                           imageWriter->fileName.getValue().getString());
    return true;
  } else {
    SoDebugError::postInfo("","%s not produced.",
                           imageWriter->fileName.getValue().getString());
    return false;
  }
}

template <class SoGuiViewer> 
inline bool GuiViewer_writeJPEG(SoGuiViewer& aThis,const std::string& aFile,int aQuality = 60) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  SoImageWriter* imageWriter = soPage->getImageWriter();
  imageWriter->fileName.setValue(aFile.c_str());
  imageWriter->format.setValue(SoImageWriter::JPEG);
  imageWriter->quality.setValue(aQuality);
  imageWriter->enable();

  SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
  viewportRegion->highlighted.setValue(FALSE);
  aThis.render();
  viewportRegion->highlighted.setValue(TRUE);

  imageWriter->disable();
  if(imageWriter->getStatus()) {
    SoDebugError::postInfo("","%s produced.",
                           imageWriter->fileName.getValue().getString());
    return true;
  } else {
    SoDebugError::postInfo("","%s not produced.",
                           imageWriter->fileName.getValue().getString());
    return false;
  }
}

template <class SoGuiViewer> 
inline bool GuiViewer_writePS(SoGuiViewer& aThis,const std::string& aFile) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  SoImageWriter* imageWriter = soPage->getImageWriter();
  imageWriter->fileName.setValue(aFile.c_str());
  imageWriter->format.setValue(SoImageWriter::POST_SCRIPT);
  imageWriter->enable();

  SoViewportRegion* viewportRegion = soRegion->getViewportRegion();
  viewportRegion->highlighted.setValue(FALSE);
  aThis.render();
  viewportRegion->highlighted.setValue(TRUE);

  imageWriter->disable();
  if(imageWriter->getStatus()) {
    SoDebugError::postInfo("","%s produced.",
                           imageWriter->fileName.getValue().getString());
    return true;
  } else {
    SoDebugError::postInfo("","%s not produced.",
                           imageWriter->fileName.getValue().getString());
    return false;
  }
}  

inline bool GuiViewer_writePSVec(
 const std::string& aFile
,const SbViewportRegion& aRegion
,const SbColor& aBackground
,SoNode* aSceneGraph) {
  // Open file :
  SbPainterPS painterPS;
  painterPS.openFileForWriting(aFile.c_str());
  // Produce view :
  SoPainterAction* action = new SoPainterAction(aRegion,&painterPS);
  action->setBackgroundColor(aBackground);
  action->apply(aSceneGraph);
  SoDebugError::postInfo("","View put in %s.",painterPS.getFileName());
  delete action;
  // Close file :
  bool status;
  if(painterPS.getStream()) {
    SoDebugError::postInfo("","%s produced.",painterPS.getFileName());
    status = true;
  } else {
    SoDebugError::postInfo("","%s not produced.",painterPS.getFileName());
    status = false;
  }
  painterPS.closeStream();
  return status;
}

template <class SoGuiViewer> 
inline bool GuiViewer_writeGL2(SoGuiViewer& aThis,
                        const std::string& aFile,
                        SoGL2PSAction::Format aFormat,
                        const std::string& aOptions) {
  SoGLRenderAction* action = aThis.getGLRenderAction();  
  if(!action) return false;
  if(action->isOfType(SoGL2PSAction::getClassTypeId())) { 
    SoGL2PSAction* gl2ps_action = (SoGL2PSAction*)action;
    gl2ps_action->setPageOptions(aOptions.c_str());
    gl2ps_action->setFileFormat(aFormat);
    gl2ps_action->setFileName(aFile.c_str());
    gl2ps_action->enableFileWriting();
    aThis.render();
    gl2ps_action->disableFileWriting();
    return true;
  } else {
    return false;
  }
}

template <class SoGuiViewer> 
inline bool GuiViewer_writeScene(SoGuiViewer& aThis,
                          const std::string& aFile,
                          const std::string& aFormat,
                          const std::string& aWhat,
                          bool aBinary,
                          bool aAltRep) {
  SoNode* node = aThis.getSceneGraph();
  if(!node) {
    SoDebugError::postInfo("GuiViewer_writeScene",
                           "No scene graph.");
    return false;
  }
  if(node->isOfType(SoPage::getClassTypeId())==TRUE) {
    //aWhat : page, region, scene staticScene dynamicScene
    SoPage* soPage = (SoPage*)node;
    if(aWhat=="page") {
      if(SoTools::write(*soPage,
                        aFile.c_str(),
                        aFormat.c_str(),
                        aBinary?TRUE:FALSE,
                        aAltRep?TRUE:FALSE)==FALSE) {
        SoDebugError::postInfo("GuiViewer_writeScene",
                               "Can't write file \"%s\".",aFile.c_str());
        return false;
      }
    } else if(aWhat=="region") {
      SoRegion* soRegion = soPage->getHighlightedRegion();
      //SoRegion* soRegion = soPage->currentRegion();
      if(!soRegion) {
        SoDebugError::postInfo("GuiViewer_writeScene",
                               "Can't get current region.");
        return false;
      }
      if(SoTools::write(*soRegion,
                        aFile.c_str(),
                        aFormat.c_str(),
                        aBinary?TRUE:FALSE,
                        aAltRep?TRUE:FALSE)==FALSE) {
        SoDebugError::postInfo("GuiViewer_writeScene",
                               "Can't write file \"%s\".",aFile.c_str());
        return false;
      }
    } else {
      SoRegion* soRegion = soPage->getHighlightedRegion();
      //SoRegion* soRegion = soPage->currentRegion();
      if(!soRegion) {
        SoDebugError::postInfo("GuiViewer_writeScene",
                               "Can't get current region.");
        return false;
      }
      SoNode* node = soRegion->getNodeOfInterest(aWhat.c_str());
      if(!node) {
        SoDebugError::postInfo("GuiViewer_writeScene",
                               "Can't get write node for \"%s\".",
                               aWhat.c_str());
        return false;
      }
      if(SoTools::write(*node,
                      aFile.c_str(),
                      aFormat.c_str(),
                      aBinary?TRUE:FALSE,
                      aAltRep?TRUE:FALSE)==FALSE) {
        SoDebugError::postInfo("GuiViewer_writeScene",
                               "Can't write file \"%s\".",aFile.c_str());
        return false;
      }
    }
  } else {
    if(SoTools::write(*node,
                      aFile.c_str(),
                      aFormat.c_str(),
                      aBinary?TRUE:FALSE,
                      aAltRep?TRUE:FALSE)==FALSE) {
      SoDebugError::postInfo("GuiViewer_writeScene",
                             "Can't write file \"%s\".",aFile.c_str());
      return false;
    }
  }
  return true;
}

template <class SoGuiViewer> 
inline bool GuiViewer_write(SoGuiViewer& aThis,
                     const std::string& aFormat,
                     const std::string& aFile,
                     const std::string& aOptions) {
    //printf("debug : GuiViewer_write : \"%s\" \"%s\"\n",
    //aFormat.c_str(),aFile.c_str());
    std::string format(aFormat);
    Lib::smanip::tolowercase(format);

    if(format=="guessed") format = Lib::smanip::suffix(aFile);

    if(format=="gif") {
      return GuiViewer_writeGIF(aThis,aFile);
    } else if(format=="jpeg") {
      int quality;
      if(!Lib::smanip::toint(aOptions,quality)) {
        SoDebugError::postInfo("GuiViewer_write",
                               "quality option \"%s\" not an int.",
                               aOptions.c_str());
        return false;
      }
      return GuiViewer_writeJPEG(aThis,aFile,quality);
    } else if(format=="ps") {
      GuiViewer_writePS(aThis,aFile);
      return true;
    } else if(format=="gl2ps") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::EPS,aOptions);
    } else if(format=="gl2tex") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::TEX,aOptions);
    } else if(format=="gl2pdf") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::PDF,aOptions);
    } else if(format=="gl2svg") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::SVG,aOptions);
    } else if(format=="gl2pgf") {
      return GuiViewer_writeGL2(aThis,aFile,SoGL2PSAction::PGF,aOptions);
    } else if(format=="psvec") {
      return GuiViewer_writePSVec(aFile,
                 aThis.getViewportRegion(),
                 aThis.getBackgroundColor(),
                 aThis.getSceneGraph());
    } else if( (format=="hiv") || (format=="iv") || (format=="inventor")  ||
               (format=="wrl") || (format=="vrml2") || 
               (format=="pov") ){

      bool binary = false;
      bool altRep = false;
      std::string what = "scene"; //scene staticScene dynamicScene region page

     {Lib::Arguments args(aOptions,";",true);
      args.find("binary",binary);
      args.find("altRep",altRep);
      std::string value;
      if(args.find("what",value)) what = value;}

      return GuiViewer_writeScene(aThis,aFile,format,what,binary,altRep);

    } else {
      SoDebugError::postInfo("GuiViewer_write",
                             "unknown format \"%s\".",
                             aFormat.c_str());
    }
    return false;
}

//template <class SoGuiViewer> 
//inline bool GuiViewer_collect(SoGuiViewer& aThis,int aX,int aY,int aW,int aH){
//  const SbViewportRegion& vpRegion = aThis.getViewportRegion();
//  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
//  if(!soPage) return false;
//  soPage->collect(aX,aY,aW,aH,vpRegion);
//  return true;
//}

template <class SoGuiViewer> 
inline bool GuiViewer_map(SoGuiViewer& aThis,
                 const SbVec2s& aPos,  // Rectangle position (GL coordinates).
                 const SbVec2s& aSize, // Rectangle size.
                 const SbVec2s& aWindow) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  soRegion->pushCamera();
  soRegion->viewRectangle(aPos,aSize,aWindow);
  return true;
}

template <class SoGuiViewer> 
inline bool GuiViewer_isInCurrentRegion(SoGuiViewer& aThis,const SbVec2s& aPos) {
  // aPos is in GL coordinates.
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  SoViewportRegion* soViewportRegion = soRegion->getViewportRegion();
  int rx,ry;
  soViewportRegion->getPositionPixels(rx,ry);
  unsigned int rw,rh;
  soViewportRegion->getSizePixels(rw,rh);
  if(aPos[0]<int(rx)) return false;
  if(aPos[0]>=int(rx+rw)) return false;
  if(aPos[1]<int(ry)) return false;
  if(aPos[1]>=int(ry+rh)) return false;
  return true;
}

template <class SoGuiViewer> 
inline bool GuiViewer_pushCamera(SoGuiViewer& aThis) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  soRegion->pushCamera();
  return true;
}

template <class SoGuiViewer> 
inline bool GuiViewer_popCamera(SoGuiViewer& aThis) {
  SoPage* soPage = GuiViewer_SoPage<SoGuiViewer>(aThis);
  if(!soPage) return false;
  SoRegion* soRegion = soPage->getHighlightedRegion();
  if(!soRegion) return false;
  soRegion->popCamera();
  return true;
}

template <class SoGuiFullViewer> 
inline bool viewer_proxy(
 void* aNativeViewer
,const std::string& aWhat
,const std::vector<std::string>& aArgs
,void*& aReturn
){
  SoGuiFullViewer* viewer = (SoGuiFullViewer*)aNativeViewer;
  aReturn = 0;

  if(aWhat=="interest") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    std::string interest = aArgs[0];

    SoNode* node = viewer->getSceneGraph();
    if(!node) {
      SoDebugError::postInfo("viewer_proxy","No scene graph.");
      return false;
    }
    if(node->isOfType(SoPage::getClassTypeId())==TRUE) {
      //interest : page, region, scene staticScene dynamicScene
      SoPage* soPage = (SoPage*)node;
      if(interest=="page") {
        aReturn = soPage;
        return true;
      } else if(interest=="region") {
        //SoRegion* soRegion = soPage->getHighlightedRegion();
        SoRegion* soRegion = soPage->currentRegion();
        if(!soRegion) {
          SoDebugError::postInfo("viewer_proxy",
                                 "Can't get region.");
          return false;
        }
        aReturn = soRegion;
        return true;
      } else {
        //SoRegion* soRegion = soPage->getHighlightedRegion();
        SoRegion* soRegion = soPage->currentRegion();
        if(!soRegion) {
          SoDebugError::postInfo("viewer_proxy",
                                 "Can't get region.");
          return false;
        }
        SoNode* node = soRegion->getNodeOfInterest(interest.c_str());
        if(!node) {
          SoDebugError::postInfo("GuiViewer_writeScene",
                                 "Can't get interest node for \"%s\".",
                                 interest.c_str());
          return false;
        }
        aReturn = node;
        return true;
      }

    } else {

      if(interest=="scene") {
        aReturn = node;
        return true;
      }

      SoDebugError::postInfo("viewer_proxy",
                             "unknown inerest keyword %s.",interest.c_str());
      return false;

    }

  } else if(aWhat=="write") {

    if(aArgs.size()!=3) {
      SoDebugError::postInfo("viewer_proxy",
                             "three arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    std::string file = aArgs[0];
    std::string format = aArgs[1];
    std::string opts = aArgs[2];

    return GuiViewer_write<SoGuiFullViewer>(*viewer,format,file,opts);

  } else if(aWhat=="read_scene") {
    if(aArgs.size()!=3) {
      SoDebugError::postInfo("viewer_proxy",
                             "three arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    std::vector<std::string> files = Lib::smanip::words(aArgs[0],"\n");
    if(files.empty()) return true;

    //std::string format = aArgs[1];
    //std::string placement = aArgs[2];

    SoNode* soNode = viewer->getSceneGraph();
    if(!soNode) return false;

    if(soNode->isOfType(SoPage::getClassTypeId())) {
      SoPage* soPage = (SoPage*)soNode;    
      SoRegion* soRegion = soPage->getHighlightedRegion();
      //SoRegion* soRegion = soPage->currentRegion();
      if(!soRegion) return false;
      if(soRegion->isOfType(SoImageRegion::getClassTypeId())) {
        SoImageRegion* soImageRegion = (SoImageRegion*)soRegion;
        HEPVis_SoImage* soImage = soImageRegion->getImage();

        for(std::vector<std::string>::iterator it = files.begin();
          it!=files.end();++it) {
          std::string format = aArgs[1];
          if(format=="guessed") format = Lib::smanip::suffix(*it);
          if(format=="gif")  {
            soImage->format.setValue(HEPVis_SoImage::GIF);
            soImage->fileName.setValue((*it).c_str());
          } else if( (format=="jpg") || (format=="jpeg"))   {
            soImage->format.setValue(HEPVis_SoImage::JPEG);
            soImage->fileName.setValue((*it).c_str());
          } else if( (format=="fits") || (format=="fit")) {
            soImage->format.setValue(HEPVis_SoImage::FITS);
            soImage->fileName.setValue((*it).c_str());
          } else {
            SoDebugError::postInfo("viewer_proxy",
                                   "image format %s not supported.",
                                   format.c_str());
          }
        }

      } else {

        //aArgs[1] not used.
        std::string placement = aArgs[2];
        for(std::vector<std::string>::iterator it = files.begin();
          it!=files.end();++it) {
          SoInput soInput;
          if(!soInput.openFile((*it).c_str())) return false;
          SoSeparator* separator = SoDB::readAll(&soInput); 
          if(!separator) return false;
          //separator ref count is 0.

          if(soRegion->doIt(SbAddNode(separator,placement.c_str()))==FALSE) {
            separator->unref();
            return false;
          }
        }

      }
    } else if(soNode->isOfType(SoGroup::getClassTypeId())) {
      //aArgs[1] not used.
      //aArgs[2] not used.
      for(std::vector<std::string>::iterator it = files.begin();
        it!=files.end();++it) {
        SoInput soInput;
        if(!soInput.openFile((*it).c_str())) return false;
        SoSeparator* separator = SoDB::readAll(&soInput); 
        if(!separator) return false;
        //separator ref count is 0.
        ((SoGroup*)soNode)->addChild(separator);      
      }
    } else {
      return false;
    }

  } else if(aWhat=="clear") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one argument expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    SoNode* soNode = viewer->getSceneGraph();
    if(!soNode) return false;
    if(soNode->isOfType(SoPage::getClassTypeId())) {
      SoPage* soPage = (SoPage*)soNode;    
      SoRegion* soRegion = soPage->getHighlightedRegion();
      //SoRegion* soRegion = soPage->currentRegion();
      if(!soRegion) return false;
      std::string what = aArgs[0];
      soRegion->clear(what.c_str());
    } else if(soNode->isOfType(SoGroup::getClassTypeId())) {
      ((SoGroup*)soNode)->removeAllChildren();      
    } else {
      return false;
    }

  } else if(aWhat=="remove_manips") {

    SoNode* soNode = viewer->getSceneGraph();
    if(!soNode) return false;
    if(soNode->isOfType(SoPage::getClassTypeId())) {
      SoPage* soPage = (SoPage*)soNode;    
      SoRegion* soRegion = soPage->getHighlightedRegion();
      //SoRegion* soRegion = soPage->currentRegion();
      if(!soRegion) return false;
      if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return false;
      SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;
      SoSeparator* sep = (SoSeparator*)displayRegion->getSceneSeparator();
      SoTools::removeManips(*sep);
    } else {
      SoTools::removeManips(*soNode);
    }

  } else if(aWhat=="detect_intersections") {

    if(aArgs.size()!=2) {
      SoDebugError::postInfo("viewer_proxy",
                             "two arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool first;
    if(!Lib::smanip::tobool(aArgs[0],first)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }
    std::string what = aArgs[1];

    SoNode* soNode = viewer->getSceneGraph();
    if(!soNode) return false;
    if(soNode->isOfType(SoPage::getClassTypeId())) {
      SoPage* soPage = (SoPage*)soNode;    
      SoRegion* soRegion = soPage->getHighlightedRegion();
      //SoRegion* soRegion = soPage->currentRegion();
      if(!soRegion) return false;
      SoNode* node = soRegion->getNodeOfInterest(what.c_str());
      if(!node) {
        SoDebugError::postInfo("GuiViewer_writeScene",
                               "Can't get node of interest \"%s\".",
                               what.c_str());
        return false;
      }
      SoTools::removeManips(*node);
      if(first) {
        SoTools::detectIntersection(*node,TRUE);
      } else {
        SoTools::detectIntersection(*node,FALSE);
      }
    } else {
      SoTools::removeManips(*soNode);
      if(first) {
        SoTools::detectIntersection(*soNode,TRUE);
      } else {
        SoTools::detectIntersection(*soNode,FALSE);
      }
    }

  } else if(aWhat=="render") {
    viewer->render();
  } else if(aWhat=="view_all") {
    viewer->viewAll();
  } else if(aWhat=="set_auto_clipping") {

    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool value;
    if(!Lib::smanip::tobool(aArgs[0],value)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }
    if(value) {
      if(viewer->isAutoClipping()==FALSE) viewer->setAutoClipping(TRUE);
    } else {
      if(viewer->isAutoClipping()==TRUE) viewer->setAutoClipping(FALSE);
    }

  } else if(aWhat=="set_decoration") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool value;
    if(!Lib::smanip::tobool(aArgs[0],value)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }

    if(value) {
      if(viewer->isDecoration()==FALSE) viewer->setDecoration(TRUE);
    } else {
      if(viewer->isDecoration()==TRUE) viewer->setDecoration(FALSE);
    }

  } else if(aWhat=="set_viewing") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool value;
    if(!Lib::smanip::tobool(aArgs[0],value)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }

    if(value) {
      if(viewer->isViewing()==FALSE) viewer->setViewing(TRUE);
    } else {
      if(viewer->isViewing()==TRUE) viewer->setViewing(FALSE);
    }

  } else if(aWhat=="set_head_light") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool value;
    if(!Lib::smanip::tobool(aArgs[0],value)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }

    if(value) {
      if(viewer->isHeadlight()==FALSE) viewer->setHeadlight(TRUE);
    } else {
      if(viewer->isHeadlight()==TRUE) viewer->setHeadlight(FALSE);
    }

  } else if(aWhat=="set_smoothing") {
    SoGLRenderAction* action = viewer->getGLRenderAction();  
    if(!action) return false;

    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool value;
    if(!Lib::smanip::tobool(aArgs[0],value)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }

    if(value) {
      if(action->isSmoothing()==FALSE) action->setSmoothing(TRUE);
    } else {
      if(action->isSmoothing()==TRUE) action->setSmoothing(FALSE);
    }

  } else if(aWhat=="set_background_color") {

    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    double r,g,b;
    if(!Lib::smanip::torgb(aArgs[0],r,g,b)) return false;
    viewer->setBackgroundColor(SbColor((float)r,(float)g,(float)b));

  } else if(aWhat=="set_title") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    viewer->setTitle(aArgs[0].c_str());

  } else if(aWhat=="viewport_region") {

    const SbViewportRegion& vpr = viewer->getViewportRegion();
    aReturn = (void*)(&vpr);

  } else if(aWhat=="getGLRenderAction") {

    SoGLRenderAction* action = viewer->getGLRenderAction();
    aReturn = (void*)action;

  } else if(aWhat=="setTransparencyType") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("setTransparencyType",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }

    int type;
    if(!Lib::smanip::toint(aArgs[0],type)) {
      SoDebugError::postInfo("setTransparencyType",
                             "option \"%s\" not an int.",
                             aArgs[0].c_str());
      return false;
    }

    viewer->setTransparencyType((SoGLRenderAction::TransparencyType)type);

  } else if(aWhat=="transparencyType") {

    SoGLRenderAction::TransparencyType type = viewer->getTransparencyType();
    int* i = new int; //should be deleted by the caller.
    *i = (int)type;
    aReturn = (void*)i;

  } else if(aWhat=="collect") {
    if(aArgs.size()!=4) {
      SoDebugError::postInfo("viewer_proxy",
                             "four arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }

    int x;
    if(!Lib::smanip::toint(aArgs[0],x)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not an int.",
                             aArgs[0].c_str());
      return false;
    }
    int y;
    if(!Lib::smanip::toint(aArgs[1],y)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not an int.",
                             aArgs[1].c_str());
      return false;
    }
    unsigned int w;
    if(!Lib::smanip::touint(aArgs[2],w)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not an unsigned int.",
                             aArgs[2].c_str());
      return false;
    }
    unsigned int h;
    if(!Lib::smanip::touint(aArgs[3],h)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not an unsigned int.",
                             aArgs[3].c_str());
      return false;
    }

    const SbViewportRegion& vpr = viewer->getViewportRegion();

    SoNode* soNode = viewer->getSceneGraph();
    if(!soNode) return false;
    if(soNode->isOfType(SoPage::getClassTypeId())) {
      SoPage* soPage = (SoPage*)soNode;    
      soPage->collect(x,y,w,h,vpr);
    }


  } else {
    SoDebugError::postInfo("viewer_proxy","bad keyword %s.",aWhat.c_str());
    return false;
  }
  return true;
}

template <class SoGuiFullViewer,class SoGuiExaminerViewer> 
inline bool viewer_proxy(
 void* aNativeViewer
,const std::string& aWhat
,const std::vector<std::string>& aArgs
,void*& aReturn
){
  aReturn = 0;

  SoGuiExaminerViewer* viewer = (SoGuiExaminerViewer*)aNativeViewer;

  if(aWhat=="set_feedback_visibility") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool value;
    if(!Lib::smanip::tobool(aArgs[0],value)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }

    if(value) {
      if(viewer->isFeedbackVisible()==FALSE) 
        viewer->setFeedbackVisibility(TRUE);
    } else {
      if(viewer->isFeedbackVisible()==TRUE) 
        viewer->setFeedbackVisibility(FALSE);
    }
    return true;

  } else if(aWhat=="set_animation") {
    if(aArgs.size()!=1) {
      SoDebugError::postInfo("viewer_proxy",
                             "one arguments expected for keyword %s.",
                             aWhat.c_str());
      return false;
    }
    bool value;
    if(!Lib::smanip::tobool(aArgs[0],value)) {
      SoDebugError::postInfo("viewer_proxy",
                             "option \"%s\" not a boolean.",
                             aArgs[0].c_str());
      return false;
    }
    if(value) {
      if(viewer->isAnimationEnabled()==FALSE) 
        viewer->setAnimationEnabled(TRUE);
    } else {
      if(viewer->isAnimationEnabled()==TRUE) 
        viewer->setAnimationEnabled(FALSE);
    }
    return true;

  } else if(aWhat=="stop_animating") {
    viewer->stopAnimating();
    return true;
  }

  return viewer_proxy<SoGuiFullViewer>(aNativeViewer,aWhat,aArgs,aReturn);
}



#endif
